-- Initialization

GearEvaluator = LibStub("AceAddon-3.0"):NewAddon("GearEvaluator", "AceEvent-3.0", "AceConsole-3.0", "AceBucket-3.0")

local slotnames = {"HeadSlot",
  "NeckSlot",
  "ShoulderSlot",
  "BackSlot",
  "ChestSlot",
  "ShirtSlot",
  "TabardSlot",
  "WristSlot",
  "HandsSlot",
  "WaistSlot",
  "LegsSlot",
  "FeetSlot",
  "Finger0Slot",
  "Finger1Slot",
  "Trinket0Slot",
  "Trinket1Slot",
  "MainHandSlot",
  "SecondaryHandSlot",
  "RangedSlot",
  "AmmoSlot",
  --"Bag0Slot",
  --"Bag1Slot",
  --"Bag2Slot",
  --"Bag3Slot"
}

-- Functions

function GearEvaluator:Debug(level, message)
  if (self.debuglevel and self.debuglevel >= level) then
    self:Print(message)
  end
end

function GearEvaluator:OnInitialize()
  if (not self.scores) then
    self.scores = {}
  end

  if (not self.errors) then
    self.errors = {}
  end

  if (not self.evaluators) then
    self.evaluators = {}
  end

  if (not self.dialogs) then
    self.dialogs = {}
  end

  self.db = LibStub("AceDB-3.0"):New("GearEvaluatorDB")

  if (not self.db.profile.evaluators) then
    self.db.profile.evaluators = {}
  end

  self:UpdateEvaluatorConfigTable()

  self.configTable.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)

  LibStub("AceConfig-3.0"):RegisterOptionsTable("GearEvaluator", self.configTable, { "gearevaluator" })
  self.dialogs.main = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("GearEvaluator", "GearEvaluator", nil, "general")
  self.dialogs.evaluators = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("GearEvaluator", "Evaluators", "GearEvaluator", "evaluators")
  self.dialogs.profiles = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("GearEvaluator", "Profiles", "GearEvaluator", "profiles")

  self.LDB = LibStub("LibDataBroker-1.1", true)

  if (self.LDB) then
    self.launcher = self.LDB:NewDataObject("GearEvaluator", {
      type = "launcher",
      icon = "Interface\\icons\\Ability_Mage_StudentOfTheMind",
      OnClick = function(clickedframe, button)
        if (button == "RightButton") then
          InterfaceOptionsFrame_OpenToCategory(self.dialogs.main)
        else
          InterfaceOptionsFrame_OpenToCategory(self.dialogs.evaluators)
        end
      end,
      OnTooltipShow = function(tt)
        tt:AddLine("Current Scores:")

        if (self.db.profile.evaluators) then
          for k in pairs(self.db.profile.evaluators) do
            tt:AddDoubleLine(k, self.scores[k] or self.errors[k])
          end
        end

        tt:AddLine("Left-click to modify evaluators")
        tt:AddLine("Right-click to edit preferences")
      end
    })
  end
end

function GearEvaluator:UpdateEvaluatorConfigTable()
  -- Delete non-existing entries
  for k, v in pairs(self.evaluatorConfigTable.args) do
    if (k ~= 'add' and not self.db.profile.evaluators[k]) then
      self.evaluatorConfigTable.args[k] = nil
    end
  end

  -- Add new entries
  for k, v in pairs(self.db.profile.evaluators) do
    self.evaluatorConfigTable.args[k] = {
      type = 'group',
      name = k,
      desc = "Options for "..k,
      args = {
        func = {
          name = "Evaluator function",
          type = 'input',
          set = 'UpdateEvaluatorFunction',
          get = 'GetEvaluatorFunction',
          multiline = true,
          width = 'full',
          order = 0
        },
        delete = {
          type = 'execute',
          name = "Delete",
          desc = "Click to delete the "..k.." evaluator",
          func = "DeleteEvaluatorFunction",
          confirm = true,
          confirmText = "You will not be able to undo this action!",
          order = 1
        },
        copy = {
          type = 'select',
          name = 'Copy',
          set = 'CopyEvaluatorFunction',
          values = function() local tbl = {} for k in pairs(self.db.profile.evaluators) do tbl[k] = k end return tbl end,
          order = 2
        },
        rename = {
          type = 'input',
          name = "Rename",
          desc = "Rename this evaluator function",
          get = function(info) return info[#info - 1] end,
          set = 'RenameEvaluatorFunction',
          order = 3
        },
      }
    }
  end
end

function GearEvaluator:OnEnable()
  self.invChangeBucket = self:RegisterBucketEvent("UNIT_INVENTORY_CHANGED", 0.1, "InventoryChangedCB")

  self:CompileAllEvaluators()
  self:UpdateScores()
end

function GearEvaluator:OnDisable()
  self:UnregisterBucket(self.invChangeBucket)
end

function GearEvaluator:AddEvaluatorFunction(info, name)
  if (not self.db.profile.evaluators[name]) then
    self:Print("Adding evaluator "..name.."!")

    self.db.profile.evaluators[name] = "return 0"
    self.evaluators[name] = self:CompileEvaluator(self.db.profile.evaluators[name])
    self:UpdateScore(name)

    self:UpdateEvaluatorConfigTable()
  else
    self:Print("Evaluator "..name.." is already defined!")
  end
end

function GearEvaluator:DeleteEvaluatorFunction(info)
  local name = info[#info - 1]

  self:Print("Deleting evaluator "..name.."!")

  self.db.profile.evaluators[name] = nil
  self.evaluators[name] = nil
  self.scores[name] = nil

  self:UpdateEvaluatorConfigTable()
end

function GearEvaluator:UpdateEvaluatorFunction(info, func)
  name = info[#info - 1]
  self.db.profile.evaluators[name] = func
  self.evaluators[name] = self:CompileEvaluator(func, name)
  self:UpdateScore(name)
end

function GearEvaluator:CopyEvaluatorFunction(info, src)
  dst = info[#info - 1]

  self:Print("Copying evaluator function from "..src.." to "..dst)

  self.db.profile.evaluators[dst] = self.db.profile.evaluators[src]
end

function GearEvaluator:RenameEvaluatorFunction(info, newname)
  name = info[#info - 1]

  if (name == newname) then
    return
  end

  self:Print("Renaming "..name.." to "..newname)

  self.db.profile.evaluators[newname] = self.db.profile.evaluators[name]
  self.db.profile.evaluators[name] = nil

  self.evaluators[newname] = self:CompileEvaluator(self.db.profile.evaluators[newname], newname)

  self:UpdateEvaluatorConfigTable()
  self:UpdateScore(newname)
end

function GearEvaluator:GetEvaluatorFunction(info)
  name = info[#info - 1]

  return self.db.profile.evaluators[name]
end

function GearEvaluator:CompileAllEvaluators()
  if (self.db.profile.evaluators) then
    for k, v in pairs(self.db.profile.evaluators) do
      self.evaluators[k] = self:CompileEvaluator(v, k)
    end
  end
end

function GearEvaluator:CompileEvaluator(text, name)
  local chunk, error = loadstring("return function(gear) "..text.." end", name)

  if (not chunk) then
    self:Print(error)
    self.errors[name] = error
  end

  return chunk()
end

function GearEvaluator:GetScore(evalname, gear)
  if (not self.evaluators[evalname]) then
    return nil, "Undefined evaluator '"..evalname.."'"
  else
    local func = self.evaluators[evalname]
    return pcall(func, gear)
  end
end

function GearEvaluator:UpdateScore(name)
  local success, score = self:GetScore(name, self:GetCurrentEquipment())

  if (not success) then
    self:Print(score)
    self.scores[name] = nil
    self.errors[name] = score
  else
    self.scores[name] = score
  end
end

function GearEvaluator:UpdateScores()
  for k, v in pairs(self.evaluators) do
    self:UpdateScore(k)
  end
end

function GearEvaluator:GetCurrentEquipment()
  local gear = {}

  for k, s in pairs(slotnames) do
    gear[s] = GetInventoryItemLink("player", GetInventorySlotInfo(s))
  end

  return gear
end

function GearEvaluator:GetActiveSlotNames()
  return slotnames
end

-- Event handlers

function GearEvaluator:InventoryChangedCB(units)
  if (units.player) then
    self:UpdateScores()
  end
end

-- Variable initalization

GearEvaluator.evaluatorConfigTable = {
  name = "Evaluators",
  type = 'group',
  handler = GearEvaluator,
  childGroups = 'tab',
  args = {
    add = {
      type = 'input',
      name = "Add evaluator",
      set = "AddEvaluatorFunction"
    }
  }
}

GearEvaluator.configTable = {
  name = "GearEvaluator",
  desc = "Evaluates your gear based on certain rules",
  type = 'group',
  args = {
    evaluators = GearEvaluator.evaluatorConfigTable,
    general = {
      type = 'group',
      name = "General",
      args = {
      }
    }
  }
}

--[===[@debug@
GearEvaluator.debuglevel = 1
--@end-debug@]===]
